// Stack operations

#define "PushByte (", byte, ")" as @" LD H,0 \n LD A,", @byte, @"\n LD L,A\n PUSH HL"
#define "PopByte (", byte, ")" as @" POP HL\n LD A,L LD (", @&byte, @"),A"
#define "PushWord (", word, ")" as @" LD HL,", @word, @"\n PUSH HL"
#define "PopWord (", word, ")" as @" POP HL\n LD (", @&word, @"),HL"

// Set, reset, or check system flags

#define "SysFlagOn (", bit, flags, ")"
	as @" SET ", @bit, @",(IY+", @flags, @")", "TRUE"
#define "SysFlagOff (", bit, flags, ")"
	as @" RES ", @bit, @",(IY+", @flags, @")", "FALSE"
#define "SysFlag (", bit, flags, ")"
	as @" BIT ", @bit, @",(IY+", @flags, @")", "ZFLAG"

// Bit-shifting

#define "SRLByte (", byte, ")"
	as @" LD A,(", @&byte, ")\n SRL A\n LD (", @&byte, @"A", "POFLAG"
#define "SLLByte (", byte, ")"
	as @" LD A,(", @&byte, ")\n SLL A\n LD (", @&byte, @"A", "POFLAG"
#define "SRAByte (", byte, ")"
	as @" LD A,(", @&byte, ")\n SRA A\n LD (", @&byte, @"A", "POFLAG"
#define "SLAByte (", byte, ")"
	as @" LD A,(", @&byte, ")\n SLA A\n LD (", @&byte, @"A", "POFLAG"
#define "RRByte (", byte, ")"
	as @" LD A,(", @&byte, ")\n RR A\n LD (", @&byte, @"A", "POFLAG"
#define "RLByte (", byte, ")"
	as @" LD A,(", @&byte, ")\n RL A\n LD (", @&byte, @"A", "POFLAG"
#define "RRCByte (", byte, ")"
	as @" LD A,(", @&byte, ")\n RRC A\n LD (", @&byte, @"A", "POFLAG"
#define "RLCByte (", byte, ")"
	as @" LD A,(", @&byte, ")\n RLC A\n LD (", @&byte, @"A", "POFLAG"

// Key-input related stuff

#ifdef TI83_PLUS
	#define "DisableOnKey ( )" as @" B_CALL($500B)"
	#define "RunIndicatorOn ( )" as @" B_CALL(_RunIndicOn)"
	#define "RunIndicatorOff ( )" as @" B_CALL(_RunIndicOff)"

	#define "InputKeyByte (", byte, ")"
		as @" B_CALL(_GetKey)\n LD (", @&byte, @"),A"
	#define "InputKeyWord (", word, ")"
		as @" B_CALL(_GetKey)\n LD (", @&word, @"),A\n LD (", @&word, @"+1),0"
	#define "GetKeyByte (", byte, ")"
		as @" B_CALL(_GetCSC)\n LD (", @&byte, @"),A"
	#define "GetKeyWord (", word, ")"
		as @" B_CALL(_GetCSC)\n LD (", @&word, @"),A\n LD (", @&word, @"+1),0"
#endif

// Functions and macros for multiplication and division. Do not use the
// void functions, they are used by the macros, and another set of functions
// below the macros use the macros. Use the macros if you have exact values
// and you want the result stored in a variable. Use the functions if you need
// to pass complex expressions or if you want to nest calls.

void H_Times_E() {
	#asm ".module H_Times_E\n LD D,0\n LD L,D\n LD B,8"
	#asm "_loop:\n ADD HL,HL\n JR NC,_skip\n ADD HL,DE"
	#asm "_skip:\n DJNZ _loop"
} // Taken from "Learn TI-83 Plus Assembly In 28 Days -- Day 15 -- Advanced math

void DE_Times_A() {
	#asm ".module DE_Times_A\n LD HL,0\n LD B,8"
	#asm "_loop:\n RRCA\n JR NC,_skip\n ADD HL,DE"
	#asm "_skip:\n SLA E\n RL D\n DJNZ _loop"
} // Taken from "Learn TI-83 Plus Assembly In 28 Days -- Day 15 -- Advanced math

void Div_HL_D() {
	#asm ".module Div_HL_D\n XOR A\n LD B,16"
	#asm "_loop:\n ADD HL,HL\n RLA\n JR C,_overflow\n CP D\n JR C,_skip"
	#asm "_overflow:\n SUB D\n INC L\n_skip:\n DJNZ _loop"
} // Taken from "Learn TI-83 Plus Assembly In 28 Days -- Day 15 -- Advanced math

#define "MultBBB (", byte1, byte2, result, ")"
	as @" LD A,", @byte1, @"\n LD H,A\n LD A,", @byte2,
	@"\n LD E,A", "H_Times_E();", @" LD A,L\n LD (", @&result, @"),A"
#define "MultBBW (", byte1, byte2, result, ")"
	as @" LD A,", @byte1, @"\n LD H,A\n LD A,", @byte2,
	@"\n LD E,A", "H_Times_E();", @" LD (", @&result, @"),HL"
#define "MultWBB (", word, byte, result, ")"
	as @" LD DE,", @word, @"\n LD A,", @byte,
	"DE_Times_A();", @" LD A,L\n LD (", @&result, @"),A"
#define "MultWBW (", word, byte, result, ")"
	as @" LD DE,", @word, @"\n LD A,", @byte,
	"DE_Times_A();", @" LD (", @&result, @"),HL"
#define "DivWBB (", word, byte, result, ")"
	as @" LD HL,", @word, @"\n LD A,", @byte, @"\n LD D,A",
	"Div_HL_D();", @" LD A,L\n LD (", @&result, @"),A"
#define "DivWBW (", word, byte, result, ")"
	as @" LD HL,", @word, @"\n LD A,", @byte, @"\n LD D,A",
	"Div_HL_D();", @" LD (", @&result, @"),HL"
#define "ModulusWB (", word, byte, result, ")"
	as @" LD HL,", @word, @"\n LD A,", @byte, @"\n LD D,A",
	"Div_HL_D();", @" LD (", @&result, @"),A"
#define "DivModWB (", word, byte, div, mod, ")"
	as @" LD HL,", @word, @"\n LD A,", @byte, @"\n LD D,A",
	"Div_HL_D();", @" LD (", @&div, @"),HL\n LD (", @&mod, @"),A"

byte MultBBB(byte A, byte B) { .MultWBB(A, B, this); }
word MultBBW(byte A, byte B) { .MultWBW(A, B, this); }
byte MultWBB(this, byte value) { .MultWBB(this, value, this); }
word MultWBW(this, byte value) { .MultWBW(this, value, this); }
byte DivWBB(this, byte value) { .DivWBB(this, value, this); }
word DivWBW(this, byte value) { .DivWBW(this, value, this); }
byte ModulusWB(this, byte value) { .ModulusWB(this, value, this); }

void DivModWB(word A, byte B, word &div, word &mod) {
	.DivModWB(A, B, A, B); *div = A; *mod = B;
} // This is messy, maybe you should not use it :-)

#eof