/*
(the) SquirrelBox compiler v1.5.
Copyright (C) 2007 Dan Cook

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

class Result { // Resulting value throughout an expression
	public int bytes;    // result type (byte or word)
	public String value; // numeric value or var address

	public Result(int b, String v) { bytes = b; value = v; }

	public boolean reference() {
		return value.charAt(0) == '(' && value.charAt(value.length()-1) == ')';
	}

	public boolean numeric() {
		try { Integer.parseInt(value); return true;
		} catch(NumberFormatException n){}
		return !reference() && !register();
	}

	private boolean register(char ch) {
		return ch == 'H' || ch == 'L' || (ch >= 'A' && ch <= 'F');
	}

	public boolean register() {
		if(value.length()>2) return false;
		if(value.length()>1) {
			if(!register(value.charAt(1))) return false;
		} return register(value.charAt(0));
	}

	private void setSize() {
		try {
			if(Math.abs(Integer.parseInt(value)) < 256) bytes = 1;
			else bytes = 2;
		} catch(NumberFormatException n) {
			if(bytes != 2) bytes = 1;
		}
	}

	public void update(boolean[] free) {
		if(!register()) return;
		free = set(free, value.charAt(0), false);
		if(value.length() > 1)
			free = set(free, value.charAt(1), false);
	}

	public boolean[] loadRegister(HLz80compiler c, boolean[] free) {
		if(register()) return free;
		boolean[] copy = new boolean[8];
		for(int i=0; i<8; i++) copy[i] = free[i];
		setSize();
		String reg = null;
		if(bytes == 2) {
			if(copy[6] && copy[7]) {
				reg = "HL"; copy[6] = false; copy[7] = false;
			} else if(copy[3] && copy[4]) { // DE is better for EX DE,HL
				reg = "DE"; copy[3] = false; copy[4] = false;
			} else if(copy[1] && copy[2]) {
				reg = "BC"; copy[1] = false; copy[2] = false;
			} else ranOutOfRegisters(c,"16",copy);
		} else {
			if(copy[0]) { reg = "A"; copy[0] = false; }
			else if(copy[1]) { reg = "B"; copy[1] = false; }
			else if(copy[2]) { reg = "C"; copy[2] = false; }
			else if(copy[3]) { reg = "D"; copy[3] = false; }
			else if(copy[4]) { reg = "E"; copy[4] = false; }
			else if(copy[6]) { reg = "H"; copy[6] = false; }
			else if(copy[7]) { reg = "L"; copy[7] = false; }
			else ranOutOfRegisters(c, "8", copy);
		} if(reference() && !reg.equals("A") && bytes != 2) {
			char temp = 0;
			if(copy[0]) temp = 'A';
			else if(copy[1]) temp = 'B';
			else if(copy[2]) temp = 'C';
			else if(copy[3]) temp = 'D';
			else if(copy[4]) temp = 'E';
			else if(copy[6]) temp = 'H';
			else if(copy[7]) temp = 'L';
			else ranOutOfRegisters(c, "8", copy);
			c.asmOutln(" LD "+temp+",A");
			c.asmOutln(" LD A,"+value);
			c.asmOutln(" LD "+reg+",A");
			c.asmOutln(" LD A,"+temp);
		} else c.asmOutln(" LD "+reg+","+value);
		value = reg; return copy;
	}

	public boolean[] toWord(HLz80compiler c, boolean[] free) {
		bytes = 2;
		if(value.length() > 1 || !register()) return free;
		String copy;
		if(value.charAt(0) == 'B' && free[2]) {
			copy = "BC"; free[2] = false;
		} else if(value.charAt(0) == 'C' && free[1]) {
			copy = "BC"; free[1] = false;
		} else if(value.charAt(0) == 'D' && free[4]) {
			copy = "DE"; free[4] = false;
		} else if(value.charAt(0) == 'E' && free[3]) {
			copy = "DE"; free[3] = false;
		} else if(value.charAt(0) == 'H' && free[7]) {
			copy = "HL"; free[7] = false;
		} else if(value.charAt(0) == 'L' && free[6]) {
			copy = "HL"; free[6] = false;
		} else {
			free = set(free, value.charAt(0), true);
			copy = unused(c, free, true);
		} bytes = 2;
		if(copy.charAt(1) != value.charAt(0))
			c.asmOutln(" LD "+copy.charAt(1)+','+value);
		c.asmOutln(" LD "+copy.charAt(0)+",0");
		value = copy; return free;
	}

	public String low() {
		try {
			return Integer.toString(Integer.parseInt(value)%256);
		} catch(NumberFormatException n){}
		if(register()) {
			if(value.length() == 2)
				return value.substring(1);
			return value;
		} if(reference()) return value;
		if(value.indexOf('+') >=0)
			return "("+value+")%256";
		return value+"%256";
	}

	public String high() {
		try {
			return Integer.toString(Integer.parseInt(value)/256);
		} catch(NumberFormatException n){}
		if(register()) {
			if(value.length() == 2)
				return value.substring(0,1);
			return "0";
		} if(reference()) {
			if(bytes != 2) return "0";
			return value.substring(0,value.length()-1)+"+1)";
		} String copy = value;
		if(value.indexOf('+') >=0)
			return "("+value+")/256";
		return value+"/256";
	}

	public String getByte(int high) {
		if(high > 0) return high();
		return low();
	}

	public static boolean[] freeList() {
		return new boolean[]{true,true,true,true,true,true,true,true};
	}

	public static boolean[] freeList(String use) {
		boolean[] free = freeList();
		for(int i=0; i<use.length(); i++) {
			switch(use.charAt(i)) {
			  case 'A': free[0] = false;
			  case 'B': free[1] = false;
			  case 'C': free[2] = false;
			  case 'D': free[3] = false;
			  case 'E': free[4] = false;
			  case 'F': free[5] = false;
			  case 'H': free[6] = false;
			  case 'L': free[7] = false;
			}
		} return free;
	}

	public static boolean[] freeList(Result other) {
		if(!other.register()) return freeList();
		return freeList(other.value);
	}

	public static void ranOutOfRegisters(
		HLz80compiler c, String b, boolean[] free
	) {
		String text =
			"Ran out of "+b+"-bit Registers to use :-(\nregisers in use: ";
		String reg = "ABCDEFHL";
		for(int i=0; i<8; i++) if(!free[i]) text += " "+reg.charAt(i);
		c.Error(text);
	}

	public static String unused(HLz80compiler c, boolean[] free, boolean word) {
		if(word) {
			if(free[6]&&free[7]) return "HL"; if(free[1]&&free[2]) return "BC";
			if(free[3]&&free[4]) return "DE"; ranOutOfRegisters(c,"16",free);
		} if(free[0]) return "A";
		if(free[1]) return "B"; if(free[2]) return "C";
		if(free[3]) return "D"; if(free[4]) return "E";
		if(free[6]) return "H"; if(free[7]) return "L";
		ranOutOfRegisters(c,"8",free); return null;
	}

	public static boolean[] set(boolean[] free, char reg, boolean state) {
		boolean[] copy = new boolean[free.length];
		for(int i=0; i<free.length; i++) copy[i] = free[i];
		if(reg == 'A') copy[0] = state; if(reg == 'B') copy[1] = state;
		if(reg == 'C') copy[2] = state; if(reg == 'D') copy[3] = state;
		if(reg == 'E') copy[4] = state; if(reg == 'F') copy[5] = state;
		if(reg == 'H') copy[6] = state; if(reg == 'L') copy[7] = state;
		return copy;
	}
}

class Context { // Code context within [nested] control-structures
	public int blocks;
	public String trail;
	public Context previous;
	public boolean returned;
	public variable[] locals;
	public int cont, brk;

	public Context() {
		blocks = 0; trail = "."; previous = null; returned = false;
		locals = new variable[0]; cont = -1; brk = -1;
	}

	private Context(String t, Context prev, int c, int b) {
		blocks = 0; trail = t; previous = prev; returned = false;
		locals = new variable[0]; cont = c; brk = b;
	}

	public Context passIf() {
		return new Context(trail+(blocks++)+'.', this, cont, brk);
	}

	public Context passLoop(int c, int b) {
		return new Context(trail+(blocks++)+'.', this, c, b);
	}

	public int offset() {
		if(previous == null) return 0;
		return previous.offset()+previous.locals.length;
	}

	public int length() {
		return offset()+locals.length;
	}

	public int find(String var) {
		int index;
		if(previous != null) {
			index = previous.find(var);
			if(index >= 0) return index;
		}
		index = utils.find(locals, var);
		if(index >= 0) return offset()+index;
		return -1;
	}

	public variable varAt(int index) {
		if(previous == null) {
			if(index >= locals.length) return null;
			return locals[index];
		}
		variable var = previous.varAt(index);
		if(var != null) return var;
		return locals[index-offset()];
	}

	public String nameAt(int index) {
		if(previous == null) {
			if(index >= locals.length) return null;
			return trail+locals[index].name;
		}
		String name = previous.nameAt(index);
		if(name != null) return name;
		return trail+locals[index-offset()].name;
	}

	public void remove(int index) {
		if(index < offset()) {
			previous.remove(index); return;
		}
		index -= offset();
		variable[] temp = new variable[locals.length-1];
		for(int i=0; i<index; i++) temp[i] = locals[i];
		for(int i=index; i<temp.length; i++) temp[i] = locals[i+1];
		locals = temp;
	}
}

public class HLz80compiler extends HLz80gui {
	public int Labels;
	public boolean usecphlde;
	public String[] asmincs;
	public variable[] globals;
	public String[] inits;
	public function[] funcs;
	public String[][] memory;

	public int NewLabel() {
		return Labels++;
	}

	public void comment(String com) {
		if(com == null) asmOutln();
		else {
			String text = asmOutput[current].getText();
			int fix = text.lastIndexOf('\n');
			if(fix < 0) fix = text.length()+1;
			else fix = text.length() - fix;
			if(fix > 16) fix = 9 + fix%8;
			asmOutln("                 ; ".substring(fix)+com);
		}
	}

	public void PostLabel(int label, String com) {
		asmOut("Label"+Integer.toString(label)+':');
		comment(com);
	}

	public int PostNewLabel(String comment) {
		PostLabel(Labels, comment);
		return Labels++;
	}

	public void jump(int label, String com) {
		asmOut(" JP Label"+label); comment(com);
	}

	public int[] typeOf(Context context, String token) { // {type, index}
		int index = utils.find(funcs, token);
		if(index >= 0) return new int[]{0,index};
		index = utils.find(globals, token);
		if(index >= 0) return new int[]{1,index};
		if(context != null) {
			index = context.find(token);
			if(index >= 0) return new int[]{2,index};
		} if(token.equals("this") && func != null) {
			if(func.ret.type.bytes > 0) return new int[]{3};
		} return new int[]{-1};
	}

	public boolean isDefined(Context context, String token) {
		return typeOf(context, token)[0] != -1;
	}

	public int endOfText() {
		return asmOutput[current].getText().length();
	}

	public void insert(int place, String text) {
		String asm = asmOutput[current].getText();
		if(!loading) asmOutput[current].setText(
			asm.substring(0,place)+text+asm.substring(place)
		);
	}

	public void use(Context context, String item) {
		if(loading) func.use(item);
		else func.unUse(context, item);
	}

	public void clearUnused(Context context) {
		if(!loading) func.clearUnused(context);
	}

	public void parseFunctionCall(Context context, function f, boolean[] free) {
		// assume that function name and "(" are already parsed
		int off = 0;
		variable arg;
		Result result;
		String reg = null;
		for(int i=0; i+off < f.args.length; i++) {
			if(i > 0) in.MatchChar(',');
			if(i == f.thisArg) {
				arg = f.ret; off = -1;
			} else arg = f.args[i+off];
			result = parseExpression(
				context, arg.type.bytes, free, false, true, f.addrs[i]
			); if( result.value.equals("A") ||
				result.value.equals("BC") || result.value.equals("DE")
			) asmOutln(" LD ("+arg.address+"),"+result.value);
			else if(result.bytes == 2) {
				if(!free[6] || !free[7]) {
					reg = result.unused(this, free, false);
					if(reg.equals("DE")) asmOutln(" EX DE,HL");
					else {
						asmOutln(" LD "+reg.charAt(0)+",H");
						asmOutln(" LD "+reg.charAt(1)+",L");
					}
				} asmOutln(" LD HL,"+result.value);
				asmOutln(" LD ("+arg.address+"),A");
				if(!free[6] || !free[7]) {
					if(reg.equals("DE")) asmOutln(" EX DE,HL");
					else {
						asmOutln(" LD H,"+reg.charAt(0));
						asmOutln(" LD L,"+reg.charAt(1));
					}
				}
			} else {
				if(!free[0]) {
					reg = result.unused(this, free, false);
					asmOutln(" LD "+reg+",A");
				} asmOutln(" LD A,"+result.value);
				asmOutln(" LD ("+arg.address+"),A");
				if(!free[0]) asmOutln(" LD A,"+reg);
			}
		} asmOutln(" CALL func_"+f.ret.name);
		use(context, f.ret.name);
		in.MatchChar(')');
	}

	private Result parseFactor(
		Context context, int bytes, boolean[] free,
		boolean nfirst, boolean nfc, boolean addrs
	) {
		int refs;
		boolean neg = in.getCharIf('-');
		boolean not = in.getCharIf('~');
		Result res = new Result(0, null);
		if(addrs) refs = 1; else refs = 0;
		while(in.next() == '*' || in.next() == '&') {
			if(in.getCharIf('*')) refs++;
			else { in.getChar(); refs--; }
		} if(in.next() == '(') {
			in.mark(); in.getChar();
			if(!in.frontChar(in.next())) in.pushBack();
			else {
				if(in.getIdent().equals("byte")) res.bytes = 1;
				else if(in.Matches("word")) res.bytes = 2;
				if(res.bytes == 0) in.pushBack();
				else { in.unMark(); in.MatchChar(')'); }
			}
		} // check parenthetical stuff SEPARATELY so both can be used 
		if(in.getCharIf('(')) {
			res = parseExpression(context, bytes, free, nfirst, true, false);
			in.MatchChar(')'); nfirst = res.register();
		} else if(in.numeric()) {
			int num = in.getSingleNum();
			if(num > 65535 || (bytes == 1 && num > 255))
				Warning("Value is too big and will be rounded: "+num);
			if(bytes == 1) num = ((num%256)+256)%256;
			else num = ((num%65536)+65536)%65536;
			res.value = Integer.toString(num);
		} else {
			refs++;
			variable var;
			int[] typeof = typeOf(context, in.getIdent());
			switch(typeof[0]) {
			  case 0:
				if(res.bytes == 0)
					res.bytes = funcs[typeof[1]].ret.type.bytes;
				if(in.getCharIf('(')) {
					if(funcs[typeof[1]].ret.type.bytes == 0) {
						if(nfc) Error(
							"You cannot call void functions in expressions",
							"call to function "+funcs[typeof[1]].ret.name
						); parseFunctionCall(context, funcs[typeof[1]], free);
						if(neg || not || refs!=1 || in.next()!=';') Error(
							"You cannot perform operations on a call to a"+
							"\nfunction that does not return a value (void)"+
							"\n(if this is not the case, then maybe there"+
							"\nis no semicolon after the function call?)",
							"call to function "+funcs[typeof[1]].ret.name
						); return null;
					} if(nfirst) Error(
						"Function calls must come first in expressions",
						"call to function "+funcs[typeof[1]].ret.name
					); parseFunctionCall(context, funcs[typeof[1]], free);
					res.value = funcs[typeof[1]].ret.address;
				} else res.value = "func_"+funcs[typeof[1]].ret.name;
				use(context, funcs[typeof[1]].ret.name);
				break;
			  case 1:
				var = globals[typeof[1]];
				if(res.bytes == 0)
					res.bytes = var.type.bytes;
				res.value = var.address;
				use(context, var.name);
				break;
			  case 2:
				var = context.varAt(typeof[1]);
				if(res.bytes == 0)
					res.bytes = var.type.bytes;
				res.value = var.address;
				use(context, context.nameAt(typeof[1]));
				break;
			  case 3:
				if(res.bytes == 0)
					res.bytes = func.ret.type.bytes;
				res.value = func.ret.address;
				break;
			  default: Error("Invalid item in expression", in.token);
			} if(refs == 0) res.bytes = 2;
		} if(refs < 0) Error(
			"You cannot address a number or another address\n"+
			"(maybe there is an inbalance of & and * operators?)"
		); String reg = null;
		if(in.getCharIf('[')) {
			String save = null;
			boolean check = res.numeric();
			if(check) {
				reg = res.value;
				save = asmOutput[current].getText();
			} free = res.loadRegister(this, free);
			Result index = parseExpression(context, 0, free, nfirst, true, false);
			if(check && index.numeric()) {
				asmOutput[current].setText(save);
				if(!index.value.equals("0")) {
					try {
						int value = Integer.parseInt(reg);
						value += Integer.parseInt(index.value);
						res.value = Integer.toString(value);
					} catch(NumberFormatException nfe) {
						res.value = reg+'+'+index.value;
					}
				} else res.value = reg;
			} else {
				if(index.bytes == 2)
					free = res.toWord(this, free);
				if(
					res.bytes == 2 && (
						res.value.equals("HL") ||
						index.reference() || index.register()
					)
				) {
					free = index.loadRegister(this, free);
					free = index.toWord(this, free);
					if(!res.value.equals("HL")) {
						if(!free[6] || !free[7]) {
							reg = index.unused(this, free, true);
							if(reg.equals("DE")) asmOutln(" EX DE,HL");
							else {
								asmOutln(" LD "+reg.charAt(0)+",H");
								asmOutln(" LD "+reg.charAt(1)+",L");
							}
						} asmOutln(" LD H,"+res.value.charAt(0));
						asmOutln(" LD L,"+res.value.charAt(1));
					} asmOutln(" ADD HL,"+index.value);
					if(!res.value.equals("HL")) {
						asmOutln(" LD "+res.value.charAt(0)+",H");
						asmOutln(" LD "+res.value.charAt(1)+",L");
						if(!free[6] || !free[7]) {
							if(reg.equals("DE")) asmOutln(" EX DE,HL");
							else {
								asmOutln(" LD H,"+reg.charAt(0));
								asmOutln(" LD L,"+reg.charAt(1));
							}
						}
					}
				} else {
					if(index.reference())
						free = index.loadRegister(this, free);
					if(!res.value.equals("A")) {
						if(!free[0]) {
							reg = Result.unused(this, free, false);
							asmOutln(" LD "+reg+",A");
						} asmOutln(" LD A,"+res.value);
					} for(int i=0; i<res.bytes; i++) {
						if(i == 1 && res.bytes == 2)
							asmOutln(" ADC A,"+index.getByte(i));
						else asmOutln(" ADD A,"+index.getByte(i));
						if(res.bytes == 2)
							asmOutln(" LD "+res.getByte(i)+",A");
					} if(!res.value.equals("A")) {
						asmOutln(" LD "+res.value+",A");
						if(!free[0]) asmOutln(" LD A,"+reg);
					}
				}
			} in.MatchChar(']');
		} in.checkForAsm(this, context);
		boolean saved = false;
		if(refs > 0) {
			if(!res.register()) {
				if(!res.reference()) {
					refs--; res.value = "("+res.value+')';
				} if(refs > 0) {
					free = res.loadRegister(this, free);
					if(in.next() == '=') refs--;
				}
			} while(refs-- > 0) {
				if(res.bytes == 1) {
					/*
					String copy;
					if(res.value.equals("B") && free[2]) {
						free[2] = false; copy = "BC";
					} else if(res.value.equals("C") && free[1]) {
						free[1] = false; copy = "BC";
					} else if(res.value.equals("D") && free[4]) {
						free[4] = false; copy = "DE";
					} else if(res.value.equals("E") && free[3]) {
						free[3] = false; copy = "DE";
					} else if(res.value.equals("H") && free[7]) {
						free[7] = false; copy = "HL";
					} else if(res.value.equals("L") && free[6]) {
						free[6] = false; copy = "HL";
					} else {
						free = Result.set(free, res.value.charAt(0), true);
						copy = Result.unused(this, free, true);
					} res.bytes = 2;
					asmOutln(" LD "+copy.charAt(1)+','+res.value);
					asmOutln(" LD "+copy.charAt(0)+",0");
					res.value = copy;
					*/
					free = res.toWord(this, free);
				} if(!free[0] && !saved) {
					reg = res.unused(this, free, false);
					asmOutln(" LD "+reg+",A"); saved = true;
				} if(!res.value.equals("HL")) {
					if(!free[0]) {
						reg = res.unused(this, free, false);
						asmOutln(" LD "+reg+",A");
					} String reg2 = Result.unused(
						this, Result.set(free, reg.charAt(0), false), false
					); asmOutln(" LD A,("+res.value+")\n LD "+reg2+",A");
					asmOutln(" INC "+res.value+"\n LD A,("+res.value+")");
					asmOutln(" LD "+res.value.charAt(0)+",A");
					asmOutln(" LD "+res.value.charAt(1)+','+reg2);
				} else asmOutln(" LD A,(HL)\n INC HL\n LD H,(HL)\n LD L,A");
			}
		} if(neg || not) {
			if(res.numeric()) {
				if(neg) {
					if(not) res.value += "+1";
					else res.value = "-("+res.value+')';
				} else res.value = "~("+res.value+')';
			} else {
				free = res.loadRegister(this, free);
				if(neg && not) asmOutln(" INC "+res.value);
				else if(res.bytes == 2) {
					if(!free[0] && !saved) {
						reg = Result.unused(this, free, false);
						asmOutln(" LD "+reg+",A"); saved = true;
					} asmOutln(" LD A,"+res.value.charAt(1));
					if(neg) asmOutln(" NEG");
					else asmOutln(" CPL");
					asmOutln(" LD "+res.value.charAt(1)+",A");
					asmOutln(" LD A,"+res.value.charAt(0)+"\n CPL");
					if(neg) asmOutln(" CCF\n ADC A,0");
					asmOutln(" LD "+res.value.charAt(0)+",A");
				} else {
					if(!res.value.equals("A")) {
						if(!free[0] && !saved) {
							reg = Result.unused(this, free, false);
							asmOutln(" LD "+reg+",A"); saved = true;
						} asmOutln(" LD A,"+res.value);
					}
					if(neg) asmOutln(" NEG");
					else asmOutln(" CPL");
					if(!res.value.equals("A"))
						asmOutln(" LD "+res.value+",A");
				}
			}
		} if(saved) asmOutln(" LD A,"+reg);
		return res;
	}

	public Result parseTerm(
		Context context, int bytes, boolean[] free,
		boolean nfirst, boolean nfc, boolean addrs
	) {
		Result result = parseFactor(context, bytes, free, nfirst, nfc, addrs);
		if(result == null) return null;
		char op;
		Result factor;
		boolean check;
		boolean[] copy;
		String save = null;
		String ins, reg = null;
		in.mark();
		do {
			in.reset();
			copy = free;
			if(in.eof()) {
				in.unMark(); break;
			} op = in.getChar();
			if(op == '&') ins = " AND ";
			else if(op == '|') ins = " OR ";
			else if(op == '^') ins = " XOR ";
			else { in.pushBack(); break; }
			check = result.numeric();
			if(check) {
				reg = result.value;
				save = asmOutput[current].getText();
			} copy = result.loadRegister(this, copy);
			factor = parseFactor(context, bytes, copy, true, true, false);
			if(check && factor.numeric()) {
				asmOutput[current].setText(save);
				if(!factor.value.equals("0")) {
					try {
						int value = Integer.parseInt(reg);
						if(op == '&') value &= Integer.parseInt(factor.value);
						else if(op == '|') value |= Integer.parseInt(factor.value);
						else value ^= Integer.parseInt(factor.value);
						result.value = Integer.toString(value);
					} catch(NumberFormatException nfe) {
						result.value = "("+reg+')'+op+'('+factor.value+')';
					}
				} else result.value = reg;
			} else {
				if(factor.bytes == 2) copy = result.toWord(this, copy);
				if(result.value.equals("A")) asmOutln(ins+factor.value);
				else {
					if(!copy[0]) {
						reg = factor.unused(this, copy, false);
						asmOutln(" LD "+reg+",A");
					} for(int i=0; i<result.value.length(); i++) {
						asmOutln(" LD A,"+result.value.charAt(i));
						asmOutln(ins+factor.getByte(i));
						asmOutln(" LD "+result.value.charAt(i)+",A");
					} if(!copy[0]) asmOutln(" LD A,"+reg);
				}
			} if(result.register()) {
				free = Result.set(free, result.value.charAt(0), false);
				if(result.value.length() > 1)
					free = Result.set(free, result.value.charAt(1), false);
			}
		} while(true);
		return result;
	}

	public Result parseExpression(
		Context context, int bytes, boolean[] free,
		boolean nfirst, boolean nfc, boolean addrs
	) {
		Result result = parseTerm(context, bytes, free, nfirst, nfc, addrs);
		if(result == null) return null;
		char op;
		String add;
		Result term;
		boolean check;
		boolean[] copy;
		String reg = null;
		String save = null;
		in.mark();
		do {
			in.reset();
			copy = free;
			if(in.eof()) {
				in.unMark(); break;
			} op = in.getChar();
			if(op == '+') add = " ADD ";
			else if(op == '-') add = " SUB ";
			else { in.pushBack(); break; }
			check = result.numeric();
			if(check) {
				reg = result.value;
				save = asmOutput[current].getText();
			} copy = result.loadRegister(this, copy);
			term = parseTerm(context, bytes, copy, true, true, false);
			if(check && term.numeric()) {
				asmOutput[current].setText(save);
				if(!term.value.equals("0")) {
					try {
						int value = Integer.parseInt(reg);
						if(op == '+') value += Integer.parseInt(term.value);
						else value -= Integer.parseInt(term.value);
						result.value = Integer.toString(value);
					} catch(NumberFormatException nfe) {
						result.value = reg+'+'+term.value;
					}
				} else result.value = reg;
			} else {
				if(term.bytes == 2)
					copy = result.toWord(this, copy);
				if(
					result.bytes == 2 && (
						result.value.equals("HL") ||
						term.reference() || term.register()
					)
				) {
					if(op == '-') add = " SBC ";
					copy = term.loadRegister(this, copy);
					copy = term.toWord(this, copy);
					if(!result.value.equals("HL")) {
						if(!copy[6] || !copy[7]) {
							reg = term.unused(this, copy, true);
							if(reg.equals("DE")) asmOutln(" EX DE,HL");
							else {
								asmOutln(" LD "+reg.charAt(0)+",H");
								asmOutln(" LD "+reg.charAt(1)+",L");
							}
						} asmOutln(" LD H,"+result.value.charAt(0));
						asmOutln(" LD L,"+result.value.charAt(1));
					} asmOutln(add+"HL,"+term.value);
					if(!result.value.equals("HL")) {
						asmOutln(" LD "+result.value.charAt(0)+",H");
						asmOutln(" LD "+result.value.charAt(1)+",L");
						if(!copy[6] || !copy[7]) {
							if(reg.equals("DE")) asmOutln(" EX DE,HL");
							else {
								asmOutln(" LD H,"+reg.charAt(0));
								asmOutln(" LD L,"+reg.charAt(1));
							}
						}
					}
				} else {
					if(term.reference())
						copy = term.loadRegister(this, copy);
					if(!result.value.equals("A")) {
						if(!copy[0]) {
							reg = Result.unused(this, copy, false);
							asmOutln(" LD "+reg+",A");
						} asmOutln(" LD A,"+result.value);
					} for(int i=0; i<result.bytes; i++) {
						if(i == 1 && result.bytes == 2) {
							if(op == '+') add = " ADC ";
							else add = " SBC ";
						} asmOut(add);
						if(!add.equals(" SUB ")) asmOut("A,");
						asmOutln(term.getByte(i));
						if(result.bytes == 2)
							asmOutln(" LD "+result.getByte(i)+",A");
					} if(!result.value.equals("A")) {
						asmOutln(" LD "+result.value+",A");
						if(!copy[0]) asmOutln(" LD A,"+reg);
					}
				}
			} if(result.register()) {
				free = Result.set(free, result.value.charAt(0), false);
				if(result.value.length() > 1)
					free = Result.set(free, result.value.charAt(1), false);
			}
		} while(true);
		return result;
	}

	public void parseCondition(
		Context context, String ins, String jpLbl, boolean swap
	) {
		String flag;
		boolean not = false;
		boolean[] free = null;
		if(in.next() == 'n') {
			in.mark();
			if(in.getIdent().equals("not")) {
				if(in.getCharIf('(')) {
					not = true; in.unMark();
				} else in.pushBack();
			} else in.pushBack();
		} else if(in.next() == '!') {
			in.mark(); in.getChar();
			if(in.getCharIf('(')) {
				in.unMark(); not = true;
			} else in.pushBack();
		} if(in.next() >= 'C' && in.next() <= 'Z') {
			in.mark(); flag = null;
			if(in.getIdent().equals("CFLAG")) flag = "C";
			else if(in.Matches("HFLAG")) flag = "H";
			else if(in.Matches("NCFLAG")) flag = "NC";
			else if(in.Matches("NHFLAG")) flag = "NH";
			else if(in.Matches("NNFLAG")) flag = "NN";
			else if(in.Matches("NFLAG")) flag = "N";
			else if(in.Matches("NSFLAG")) flag = "NS";
			else if(in.Matches("NZFLAG")) flag = "NZ";
			else if(in.Matches("PEFLAG") || in.Matches("POFLAG")) {
				if(in.Matches("PEFLAG")^swap) flag = "PO";
				else flag = "PO";
			} else if(in.Matches("SFLAG")) flag = "S";
			else if(in.Matches("ZFLAG")) flag = "Z";
			else if(in.Matches("TRUE") || in.Matches("FALSE")) {
				if(in.Matches("FALSE")^swap) {
					if(jpLbl == null) asmOutln(ins);
					else asmOutln(ins+jpLbl);
				} return;
			} if(flag != null) {
				in.unMark();
				if(not) in.MatchChar(')');
				if(swap == not) {
					if(flag.length() > 1) flag = flag.substring(1);
					else if(flag.charAt(0) != 'P') flag = "N"+flag;
				}
				if(jpLbl == null) asmOutln(ins+flag);
				else asmOutln(ins+flag+','+jpLbl);
				return;
			} in.pushBack();
		} Result lhs = parseExpression(
			context, 0, Result.freeList(), false, true, false
		); char op = in.get();
		switch(op) {
		  case '=':
			in.MatchChar('='); op = 0; break;   // ==
		  case '!':
			if(in.getCharIf('>')) op = 4;       // !<
			else if(in.getCharIf('<')) op = 2;  // !>
			else { in.MatchChar('='); op = 1; } // !=
			break;
		  case '>':
		  	if(in.getCharIf('=')) op = 2; // >=
		  	else if(swap) op = 5;         // >
		  	else op = 4; break;           // <=
		  case '<':
		  	if(in.getCharIf('=')) {
		  		if(swap) op = 4;  // <=
		  		else op = 5;      // >
		  	} else op = 3; break; // <
		  default: Error(
				"Relational operator expected (next char is \'"+op+"\')"
			);
		} Result rhs, temp;
		rhs = parseExpression(
			context, lhs.bytes, Result.freeList(lhs), false, true, false
		); if(not) in.MatchChar(')');
		if(lhs.numeric() && rhs.numeric()) {
			int a = Integer.parseInt(lhs.value);
			int b = Integer.parseInt(rhs.value);
			if( (   (op==0 && a==b) || (op==3 && a<b) || (op==5 && a>b) ||
					(op==1 && a!=b) || (op==4 && a<=b) || (op==2 && a>=b)
				) == swap
			) {
				if(jpLbl == null) asmOutln(ins);
				else asmOutln(ins+jpLbl);
			}
		}
		if(rhs.value.equals("HL") || rhs.value.equals("A") && lhs.bytes != 2) {
			if(op > 1) {
				if(op < 4) op += 2;
				else op -= 2;
			} temp = rhs;
			rhs = lhs;
			lhs = temp;
		} int mid;
		if(op < 2) flag = "Z"; else flag = "C";
		if(swap^(op==0 || op==3)) flag = "N"+flag;
		free = lhs.loadRegister(this, Result.freeList(rhs));
		if(lhs.bytes == 2 || rhs.bytes == 2) {
			rhs.loadRegister(this, free);
			asmOutln(";THE FOLLOWING BETTER SAY HL: "+lhs.value);
			asmOutln(";THE FOLLOWING BETTER SAY DE: "+rhs.value);
			asmOutln(";IF NOT, REPORT IT AS AN ERROR");
			asmOutln("CALL cphlde");
		} else {
			asmOutln(";THE FOLLOWING BETTER SAY A: "+lhs.value);
			asmOutln(";IF NOT, REPORT IT AS AN ERROR");
			if(rhs.register() || !rhs.reference())
				asmOutln(" CP "+rhs.value);
			else {
				String reg = null;
				if(!free[6] || !free[7]) {
					reg = rhs.unused(this, free, true);
					if(reg.equals("DE")) asmOutln(" EX DE,HL");
					else {
						asmOutln(" LD "+reg.charAt(0)+",H");
						asmOutln(" LD "+reg.charAt(1)+",L");
					}
				} asmOutln(
					" LD HL,"+rhs.value.substring(1,rhs.value.length()-1)
				); asmOutln(" CP (HL)");
				if(!free[6] || !free[7]) {
					if(reg.equals("DE")) asmOutln(" EX DE,HL");
					else {
						asmOutln(" LD H,"+reg.charAt(0));
						asmOutln(" LD L,"+reg.charAt(1));
					}
				}
			}
		} if(op == 5) {
			mid = NewLabel();
			asmOutln(" JP C,Label"+mid);
			if(jpLbl == null) asmOutln(ins+"NZ");
			else asmOutln(ins+"NZ,"+jpLbl);
			PostLabel(mid,"mid-conditional-pass/fail");
		} if(op == 4) {
			if(jpLbl == null) {
				asmOutln(ins+'C');
				asmOutln(ins+'Z');
			} else {
				asmOutln(ins+"C,"+jpLbl);
				asmOutln(ins+"Z,"+jpLbl);
			}
		} else if(jpLbl == null) asmOutln(ins+flag);
		else asmOutln(ins+flag+','+jpLbl);
	}

	public void parseGoto(Context context) {
		String label = in.getIdent();
		if(utils.find(func.labels, in.token) == -1) {
			if(utils.find(func.newLabels, in.token) == -1)
				func.newLabels = utils.add(func.newLabels, in.token);
		} label = "Lbl_"+func.ret.name+'_'+label;
		if(in.frontChar(in.next())) {
			in.mark();
			if(in.getIdent().equals("if")) {
				in.unMark(); in.MatchChar('(');
				parseCondition(context, " JP ", label, true);
				in.MatchChar(')'); in.MatchChar(';'); return;
			} in.pushBack();
		} asmOutln(" JP "+label);
		in.MatchChar(';');
	}

	public void parseIf(Context context) {
		in.MatchChar('(');
		comment("<IF>");
		int L1 = NewLabel();
		parseCondition(context, " JP ", "Label"+L1, false);
		in.MatchChar(')');
		if(parseBlock(context.passIf(), true)) {
			int L2 = NewLabel();
			jump(L2,"skip else code");
			PostLabel(L1, "<ELSE>");
			parseBlock(context.passIf(), false);
			PostLabel(L2,"<End IF-ELSE>");
			return;
		} else PostLabel(L1,"<End IF>");
	}

	public void parseWhile(Context context, boolean until) {
		String which;
		if(until) which = "UNTIL";
		else which = "WHILE";
		in.MatchChar('(');
		int L1 = PostNewLabel("<"+which+'>');
		int L2 = NewLabel();
		parseCondition(context, " JP ", "Label"+L2, until);
		in.MatchChar(')');
		parseBlock(context.passLoop(L1, L2), false);
		jump(L1, "continue "+which);
		PostLabel(L2, "<End "+which+'>');
	}

	public void parseDo(Context context) {
		String which = null;
		int L1 = PostNewLabel("<DO->");
		int top = endOfText()-2;
		int cond = NewLabel();
		int L2 = NewLabel();
		parseBlock(context.passLoop(cond, L2), false);
		if(in.getIdent().equals("until")) which = "UNTIL";
		else if(in.token.equals("while")) which = "WHILE";
		else Error(
			"Missing condition after do (while or until expected)", in.token
		);
		in.MatchChar('(');
		parseCondition(context, " JP ", "Label"+L1, which.equals("WHILE"));
		in.MatchChar(')');
		insert(top, which);
		PostLabel(L2,"<End "+which+'>');
	}

	public void parseLoop(Context context) {
		int L1 = PostNewLabel("<LOOP>");
		int L2 = NewLabel();
		parseBlock(context.passLoop(L1, L2), false);
		jump(L1,"continue LOOP");
		PostLabel(L2,"<End LOOP>");
	}

	public void parseAssignment(Context context, Result[] lhs) {
		boolean[] free = Result.freeList();
		if(lhs == null) {
			lhs = new Result[] {
				parseExpression(context, 0, free, false, false, false)
			}; if(lhs[0] == null) return;
			free = Result.freeList(lhs[0]);
		} else {
			for(int i=0; i<lhs.length; i++) {
				boolean[] copy = Result.freeList(lhs[i]);
				for(int j=0; j<8; j++) free[j] &= copy[j];
			}
		} if(!in.getCharIf('=')) return;
		String reg = null;
		boolean saved = false;
		Result rhs = parseExpression(
			context, lhs[0].bytes, free, false, true, false
		); if(in.next() == '=') {
			Result[] copy = new Result[lhs.length+1];
			for(int i=0; i<lhs.length; i++) copy[i+1] = lhs[i];
			copy[0] = rhs; parseAssignment(context, copy); return;
		} for(int i=0; i<lhs.length; i++) {
			if(lhs[i].register()) {
				if(lhs[i].value.length() == 1) {
					/*
					String copy;
					if(lhs[i].value.equals("B") && free[2]) {
						copy = "BC"; free[2] = false;
					} else if(lhs[i].value.equals("C") && free[1]) {
						copy = "BC"; free[1] = false;
					} else if(lhs[i].value.equals("D") && free[4]) {
						copy = "DE"; free[4] = false;
					} else if(lhs[i].value.equals("E") && free[3]) {
						copy = "DE"; free[3] = false;
					} else if(lhs[i].value.equals("H") && free[7]) {
						copy = "HL"; free[7] = false;
					} else if(lhs[i].value.equals("L") && free[6]) {
						copy = "HL"; free[6] = false;
					} else {
						free = Result.set(free, lhs[i].value.charAt(0), true);
						copy = Result.unused(this, free, true);
					} lhs[i].bytes = 2;
					if(copy.charAt(1) != lhs[i].value.charAt(0))
						asmOutln(" LD "+copy.charAt(1)+','+lhs[i].value);
					asmOutln(" LD "+copy.charAt(0)+",0");
					lhs[i].value = copy;
					*/
					free = lhs[i].toWord(this, free);
				} if(lhs[i].value.equals("HL")) {
					if(!rhs.numeric())
						rhs.loadRegister(this, free);
					asmOutln(" LD (HL),"+rhs.low());
					if(lhs[i].bytes == 2)
						asmOutln(" INC HL\n LD (HL),"+rhs.high());
				} else {
					if(!free[0]) {
						if(!saved) {
							reg = Result.unused(this, free, false);
							asmOutln(" LD "+reg+",A");
						} asmOutln(" LD A,"+rhs.low());
					} asmOutln(" LD ("+lhs[i].value+"),A");
					if(lhs[i].bytes == 2) {
						asmOutln(" INC "+lhs[i].value+"\n LD A,"+rhs.high());
						asmOutln(" LD ("+lhs[i].value+"),A");
					}
				}
			} else if(lhs[i].reference()) {
				rhs.loadRegister(this, free);
				if(lhs[i].bytes == 1) {
					if(!free[0]) {
						if(!saved) {
							reg = Result.unused(this, free, false);
							asmOutln(" LD "+reg+",A");
						} asmOutln(" LD A,"+rhs.value);
					} asmOutln(" LD "+lhs[i].value+",A");
				} else asmOutln(" LD "+lhs[i].value+','+rhs.value);
			} else Error(
				"You cannot assign a value to a number (dereference it)",
				lhs[i].value
			);
		} if(saved) asmOutln(" LD A,"+reg);
	}

	public boolean parseBlock(Context context, boolean If) {
		boolean end = false;
		boolean unmark;
		context.returned = false;
		boolean bracket = in.getCharIf('{');
		if(!bracket && in.frontChar(in.next())) {
			in.mark(); in.getIdent();
			end = If == in.Matches("then") && If != in.Matches("begin");
			if(end) in.unMark();
			else in.pushBack();
		} do {
			do {
				if(in.getCharIf(';')) {
					if(!bracket && !end) {
						clearUnused(context.previous);
						if(If) {
							if(!in.eof()) {
								if(in.frontChar(in.next())) {
									in.mark();
									if(in.getIdent().equals("else")) {
										in.unMark(); return true;
									} in.pushBack();
								}
							} return false;
						} return false;
					} continue;
				} if(in.frontChar(in.next())) {
					in.mark(); if(
						in.getIdent().equals("TRUE") || in.Matches("FALSE") ||
						in.Matches("SFLAG") || in.Matches("NSFLAG") ||
						in.Matches("ZFLAG") || in.Matches("NZFLAG") ||
						in.Matches("HFLAG") || in.Matches("NHFLAG") ||
						in.Matches("PEFLAG") || in.Matches("POSFLAG") ||
						in.Matches("NFLAG") || in.Matches("NNFLAG") ||
						in.Matches("CFLAG") || in.Matches("NCFLAG")
					) { in.unMark(); continue; } in.pushBack();
				} if(in.checkForAsm(this, context)) continue;
				break;
			} while(true);
			if(bracket) {
				if(in.getCharIf('}')) {
					clearUnused(context.previous);
					if(If) {
						if(!in.eof()) {
							if(in.frontChar(in.next())) {
								in.mark();
								if(in.getIdent().equals("else")) {
									in.unMark(); return true;
								} in.pushBack();
							}
						} return false;
					} return false;
				}
			} in.mark();
			unmark = true;
			in.getToken(true);
			in.checkForAsm(this, context);
			if(in.Matches("end")) {
				if(bracket) Error("\'end\' found where '}' was expected");
				if(!end)
					Error("end found where a statement was expected");
				clearUnused(context.previous);
				return false;
			}
			if(in.Matches("else")) {
				clearUnused(context.previous);
				if(If) return true;
				Error("\"else\" found without matching \"if\"");
			}
			if(context.returned && loading)
				Warning("Seemingly unreachable statements found after return");
			context.returned = false;
			if(in.Matches("if")) parseIf(context);
			else if(in.Matches("while") || in.Matches("until")) {
				parseWhile(context, in.Matches("until"));
			}
			else if(in.Matches("do")) parseDo(context);
			else if(in.Matches("loop")) parseLoop(context);
			else if(in.Matches("goto")) parseGoto(context);
			else if(in.Matches("label")) {
				in.getIdent();
				if(loading) {
					if(utils.find(func.labels, in.token) >= 0) Error(
						"Duplicate \""+in.token+"\" label in function "+
						func.ret.name
					); func.labels = utils.add(func.labels, in.token);
				} asmOutln("Lbl_"+func.ret.name+'_'+in.token+':');
				in.MatchChar(';');
			} else if(in.Matches("continue")) {
				if(context.cont < 0) Error("No loop to continue");
				boolean nif = true;
				if(in.frontChar(in.next())) {
					in.mark();
					if(in.getIdent().equals("if")) {
						nif = false;
						in.unMark(); in.MatchChar('(');
						parseCondition(
							context, " JP ", "Label"+context.cont, true
						); in.MatchChar(')');
					} else in.pushBack();
				} if(nif) jump(context.cont, "continue");
				in.MatchChar(';');
			} else if(in.Matches("break")) {
				if(context.brk < 0) Error("No loop to break from");
				boolean nif = true;
				if(in.frontChar(in.next())) {
					in.mark();
					if(in.getIdent().equals("if")) {
						nif = false;
						in.unMark(); in.MatchChar('(');
						parseCondition(
							context, " JP ", "Label"+context.brk, true
						); in.MatchChar(')');
					} else in.pushBack();
				} if(nif) jump(context.brk, "break");
				in.MatchChar(';');
			} else if(in.Matches("return")) {
				boolean nif = true;
				if(in.frontChar(in.next())) {
					in.mark();
					if(in.getIdent().equals("if")) {
						nif = false; in.unMark();
						in.MatchChar('(');
						parseCondition(context, " RET ", null, true);
						in.MatchChar(')');
					} else in.pushBack();
				} if(nif) asmOutln(" RET");
				in.MatchChar(';'); context.returned = true;
			} else {
				String ident = in.token;
				if(Type.valid(ident, false)) {
					variable var = null;
					int[] typeof = null;
					Type type = in.getType(false);
					do {
						ident = in.getIdent();
						typeof = typeOf(context, ident);
						switch(typeof[0]) {
						  case 0: Error(
								"\""+ident +
								"\" is already defined as a function",
								funcs[typeof[1]].toString()
							);
						  case 1: Error(
								"\""+ident +
								"\" is already defined as a global variable",
								globals[typeof[1]].toString()
							);
						  case 2: Error(
								"\""+ident +
								"\" is already defined as a local variable",
								context.varAt(typeof[1]).toString()
							);
						  case 3: Error("You cannot name a variable \"this\"");
						}
						String address = "$0000";
						if(!loading) address =
							"local_vars+"+newAddress(context,type.size());
						var = new variable(type, ident, address);
						context.locals = utils.add(context.locals, var);
						if(in.next() == '(') Error(
							"Syntax Error", type.toString()+ident+'('
						);
						if(in.next() == '=') parseAssignment(
							context, new Result[] {
								new Result(type.bytes, "("+address+')')
							}
						);
					} while(in.getCharIf(','));
				} else {
					unmark = false; in.pushBack();
					parseAssignment(context, null);
				} in.MatchChar(';');
			} if(unmark) in.unMark();
		} while(bracket || end);
		clearUnused(context.previous);
		if(If) {
			if(!in.eof()) {
				if(in.frontChar(in.next())) {
					in.mark();
					if(in.getIdent().equals("else")) {
						in.unMark(); return true;
					} in.pushBack();
				}
			} return false;
		} return false;
	}

	public boolean available(Context context, int address) {
		if(memory[address].length == 0) return true;
		int index, size, i;
		for(i=0; i<memory[address].length; i++) {
			index = utils.find(funcs, memory[address][i]);
			if(index >= 0) {
				if(
					funcs[index] != func && (
						utils.find(funcs[index].used, func.ret.name) >= 0 ||
						utils.find(func.used, funcs[index].ret.name) >= 0
					)
				) return false;
			}
		}
		if(utils.find(memory[address], func.ret.name) < 0) return true;
		size = context.length();
		for(i=0; i<size; i++) {
			if(
				(context.varAt(i).getAddress() <= address) && (
					context.varAt(i).getAddress() >
					address-context.varAt(i).type.size()
				)
			) return false;
		} return true;
	}

	public int newAddress(Context context, int bytes) {
		int j, index;
		boolean pass;
		for(int i=0; i<memory.length-bytes+1; i++) {
			pass = true;
			for(j=0; j<bytes && pass; j++)
				pass = available(context, i+j);
			if(pass) {
				for(j=0; j<bytes && pass; j++) {
					if(utils.find(memory[i+j], func.ret.name) < 0)
						memory[i+j] = utils.add(memory[i+j], func.ret.name);
				} return i;
			}
		} Error("Ran out of usable runtime memory (1000 bytes wasn't enough?)");
		return -1;
	}

	private int nextValue() {
		String str;
		int size = 0;
		if(in.next() == '\"') {
			str = in.getString(true);
			size = str.length()-2;
			char ch;
			for(int i=1; i<str.length()-1; i++) {
				ch = str.charAt(i);
				if(ch=='\n' || ch=='\t' || ch=='\0') {
					if(i < str.length()-2) {
						str = str.substring(0,i)+"\","+(byte)ch +
							",\""+str.substring(i+1);
						i += 4;
					} else {
						str = str.substring(0,i)+"\","+(byte)ch;
						i += 2;
					}
				}
			}
		} else {
			str = null;
			int value = in.getWholeNum();
			do {
				size++;
				if(str == null) str = Integer.toString(value%256);
				else str += ","+(value%256); // z80 is LITTLE-ENDIAN
				value /= 256;
			} while(value > 0);
		}
		in.token = str;
		return size;
	}

	public String initGlobal(Type type) {
		if(type.dim == 0) {
			int size = nextValue();
			if(size > type.bytes) Error(
				"Value too big for datatype ("+type.toString()+')',
				in.token
			);
			return "0,".substring(0,2*(type.bytes-size))+in.token;
		}
		int i = 0;
		String init = "";
		in.MatchChar('{');
		while(!in.getCharIf('}')) {
			if(i > 0) {
				in.MatchChar(',');
				init += ',';
			}
			i += nextValue();
			init += in.token;
		}
		if(i != type.dim) Error(
			"Incorrect amount of bytes in array/n(" +
			type.dim+"bytes expected, "+i+" bytes found)",
			"{"+init+'}'
		);
		return init;
	}

	public String parseSource() {
		Type type;
		String str;
		int[] sizes;
		int[] typeof;
		variable var;
		function other;
		variable[] args;
		boolean[] addrs;
		Context context;
		String header = "";
		while(!in.eof()) {
			in.checkForAsm(null, null);
			if(in.getIf('#')) {
				if(in.getIdent().equals("asminc")) {
					str = in.getString(false);
					if(utils.find(asmincs, str) == -1)
						asmincs = utils.add(asmincs, str);
				}
				else if(in.Matches("header"))
					header += in.getString(false)+'\n';
				else Error("Invalid command here: #"+in.token);
			} else { // Do NOT "use" anything here; things get used in functions
				in.getIdent();
				if(!Type.valid(in.token, true)) Error(
					"Global variable or function expected", in.token
				); var = new variable(in.getType(false), in.getIdent(), null);
				typeof = typeOf(null, var.name);
				if(typeof[0] != 0) other = null;
				else other = funcs[typeof[1]];
				if(typeof[0] == 3) Error(
					"You cannot name a variable or function \"this\""
				); if(typeof[0] > 0) Error(
					var.name+" is already defined as a global variable"
				);
				if(in.getCharIf('(')) {
					int size;
					args = new variable[0];
					addrs = new boolean[0];
					int thisArg = -1;
					if(in.next() != ')') {
						do {
							size = addrs.length;
							boolean[] copy = new boolean[size+1];
							for(int i=0; i<size; i++) {
								copy[i] = addrs[i];
							} addrs = copy;
							if(in.getIdent().equals("this")) {
								addrs[size] = in.getCharIf('&');
								if(thisArg >= 0) Error(
									"More than one \"this\" argument"+
									" was found in function "+var.name
								); thisArg = args.length;
							} else {
								type = in.getType(false);
								addrs[size] = in.getCharIf('&');
								args = utils.add(
									args, new variable(
										type, in.getIdent(), "$0000"
									)
								); typeof = typeOf(null, in.token);
								if(typeof[0] == 0) Error(
									in.token+" is already defined as a function"
								); if(typeof[0] == 1) Error(
									in.token+
									" is already defined as a global variable"
								);
							}
						} while(in.getChar() == ',');
					} else in.MatchChar(')');
					func = new function(var, args, addrs);
					func.thisArg = thisArg;
					if(in.getCharIf(';')) {
						if(other != null) Error(
							var.name+" is already defined as a function"
						); func.code = null; funcs = utils.add(funcs, func);
					} else {
						if(other != null) {
							if(other.code != null) Error(
								var.name+" is already defined as a function"
							); if(
								other.args.length != args.length ||
								other.addrs.length != addrs.length
							) Error(
								"Function "+var.name+
								" does not match its header"
							); for(int i=0; i<args.length; i++) {
								if(
									!args[i].type.equals(other.args[i].type)
								) Error(
									"Function "+var.name +
									" does not match its header"
								);
								if(
									!args[i].name.equals(other.args[i].name)
								) Warning(
									"Inconsistant name for argument #"+i+
									" in function "+var.name+
									".\nUse the latter one ("+args[i].name+
									")?"
								);
							} for(int i=0; i<addrs.length; i++) {
								if(addrs[i] != other.addrs[i]) Warning(
									"Inconsistant use of addressing (&) for "+
									"argument #"+i+"\nin function "+var.name+
									". Use the latter one ("+addrs[i]+")?"
								);
							} other = func;
						} else funcs = utils.add(funcs, func);
						in.clear();
						context = new Context();
						context.locals = args;
						for(int j=0; j<func.args.length; j++) {
							if(args[j] != null) {
								func.args[j].address = "local_vars+"+newAddress(
									context, func.args[j].type.size()
								);
							}
						} if(func.ret.type.bytes > 0) {
							func.ret.address = "local_vars+"+newAddress(
								context, func.ret.type.size()
							);
						} parseBlock(context, false);
						func.code = in.output; in.clear();
					} for(int j=0; j<func.newLabels.length; j++) {
						if(
							utils.find(func.labels, func.newLabels[j]) == -1
						) Error(
							"Attempted to Goto a non-existant label\n" +
							"(\"Goto "+func.newLabels[j]+"\")"
						);
					} func = null;
				} else {
					if(other != null) Error(
						var.name+" is already defined as a function"
					); boolean repeat;
					do {
						if(var.type.bytes == 0)
							Error("Variables cannot be void", var.toString());
						if(!in.eof()) {
							if(in.getCharIf('@')) {
								str = null; var.address = in.getString(false);
							} else {
								var.address = "global_"+var.name;
								if(in.getCharIf('=')) str = initGlobal(var.type);
								else str = "0,0".substring(
									0, 2*(var.type.size())-1
								);
							}
						} else {
							var.address = "global_"+var.name;
							str = "0,0".substring(0,1+2*(var.type.dim-1));
						} inits = utils.add(inits, str);
						globals = utils.add(globals, var);
						repeat = in.getCharIf(',');
						if(repeat) {
							var = new variable(var.type, null, "$0000");
							var.name = in.getIdent();
							typeof = typeOf(null, var.name);
							if(typeof[0] == 0) {
								if(funcs[typeof[1]].code != null) Error(
									var.name+" is already defined as a function"
								);
							} if(typeof[0] > 0) Error(
								var.name+
								" is already defined as a global variable"
							);
						}
					} while(repeat);
					in.MatchChar(';');
				}
			}
		} return header;
	}

	private boolean[][] removeUnused(boolean[][] used, int f) {
		int index;
		for(int i=0; i<funcs[f].used.length; i++) {
			index = utils.find(globals, funcs[f].used[i]);
			if(index >= 0) used[1][index] = true;
			else {
				index = utils.find(funcs, funcs[f].used[i]);
				if(index >= 0) {
					if(!used[0][index]) {
						used[0][index] = true;
						used = removeUnused(used, index);
					}
				}
			}
		} return used;
	}

	public void removeUnused() {
		boolean[][] used = new boolean[][] {
			new boolean[funcs.length],
			new boolean[globals.length]
		};
		int index = utils.find(funcs, "main");
		if(index < 0) Error("No main function was found");
		used[0][index] = true;
		used = removeUnused(used, index);
		boolean[] temp;
		function[] tfunc;
		for(int i=0; i<used[0].length; i++) {
			if(!used[0][i]) {
				temp = new boolean[used[0].length-1];
				tfunc = new function[funcs.length-1];
				for(int j=0; j<i; j++) {
					temp[j] = used[0][j];
					tfunc[j] = funcs[j];
				}
				for(int j=i; j<used[0].length-1; j++) {
					temp[j] = used[0][j+1];
					tfunc[j] = funcs[j+1];
				}
				used[0] = temp;
				funcs = tfunc;
				i--;
			}
		} String[] tinits;
		variable[] tglob;
		for(int i=0; i<used[1].length; i++) {
			if(!used[1][i]) {
				temp = new boolean[used[1].length-1];
				tinits = new String[inits.length-1];
				tglob = new variable[globals.length-1];
				for(int j=0; j<i; j++) {
					temp[j] = used[1][j];
					tinits[j] = inits[j];
					tglob[j] = globals[j];
				}
				for(int j=i; j<used[1].length-1; j++) {
					temp[j] = used[1][j+1];
					tinits[j] = inits[j+1];
					tglob[j] = globals[j+1];
				} used[1] = temp;
				inits = tinits;
				globals = tglob;
				i--;
			}
		}
	}

	public void compile() {
		Context context;
		String text, arg;
		boolean none = true;
		for(current=0; current<asmUsed.length; current++) {
			usecphlde = false;
			if(!asmUsed[current]) continue;
			show(current, true);
			none = false;
			Labels = 0;
			inits = new String[0];
			asmincs = new String[0];
			funcs = new function[0];
			globals = new variable[0];
			memory = new String[1000][0];
			in = new Parser(code[current], this);
			if(in.eof()) {
				Message(
					"There was nothing left to compile for the "+
					asmNames[current]+" after\npreprocessing, so no [new]"+
					" code will be generated for it"
				); continue;
			} try {
				loading = true;
				String asm = parseSource();
				loading = false;
				removeUnused();
				for(int i=0; i<asmincs.length; i++)
					asmOutln("#include \""+asmincs[i]+'\"');
				if(asm.length() > 0) asmOut(asm);
				int size = utils.find(funcs, "main");
				if(size < 0) Error("No main function was found");
				if(size != 0) {
					func = funcs[0];
					funcs[0] = funcs[size];
					funcs[size] = func;
				} if(funcs[0].ret.type.bytes > 0) Error(
					"The main function must have a void return type"
				); for(int i=0; i<funcs.length; i++) {
					func = funcs[i];
					if(func.code == null) Error(
						"Unused function header:\n"+func
					); context = new Context();
					context.locals = func.args;
					in = new Parser(func.code, this);
					asmOutln("\nfunc_"+func.ret.name+':');
					parseBlock(context, false);
					if(!context.returned) asmOutln(" RET");
				} if(usecphlde) {
					asmOutln("\ncphlde:");
					asmOutln(" PUSH HL");
					asmOutln(" OR A");
					asmOutln(" SBC HL,DE");
					asmOutln(" POP HL");
					asmOutln(" RET");
				} for(int i=0; i<globals.length; i++) {
					if(inits[i] == null) continue;
					asmOut("\n"+globals[i].address+':');
					comment(globals[i].toString());
					asmOutln(" .DB "+inits[i]);
				} size = 0;
				for(int i=0; i < memory.length && memory[i].length > 0; i++) {
					for(int j=0; j<memory[i].length; j++) {
						if(isDefined(null, memory[i][j])) {
							size = i+1; break;
						}
					}
				} if(size > 0) {
					asmOutln("\nlocal_vars:");
					for(int i=0; i<size; i++) {
						asmOut(" .DB ");
						for(int j=0; j<25 && i+j<size; j++) {
							if(j!=0) asmOut(","); asmOut("0");
						} if(i+25 >= size) break; asmOutln();
					}
				} asmOutln("\n.END\n.END");
			} catch(CompileError c) {
				if(asmOutput[current].getText().length() > 0) {
					asmOutput[current].setText(
						"; ERRORS WERE FOUND IN THIS " +
						"PROGRAM AND IT IS NOT SAFE\n" +
						"; TO RUN, SO DO NOT ATTEMPT TO " +
						"COMPILE AND RUN THIS PROGRAM.\n\n" +
						asmOutput[current].getText()+"\n\n" +
						"; COMPILATION STOPPED HERE.\n" +
						"; THIS PROGRAM IS INCOMPLETE " +
						"BECAUSE THE SOURCE HAD ERRORS"
					);
				}
				String stop = "Stop compilating";
				if(
					javax.swing.JOptionPane.OK_OPTION ==
					javax.swing.JOptionPane.showOptionDialog(
						root,
						"Compilation for the "+asmNames[current] +
						" has been\nstopped because errors were found",
						"SquirrelBox v1.4",
						javax.swing.JOptionPane.OK_CANCEL_OPTION,
						javax.swing.JOptionPane.WARNING_MESSAGE,
						icon, new String[]{stop,"Next platform"}, stop
					)
				) break;
			}
		} if(none) Message(
			"No platform was defined, so there was nothing to\n" +
			"compile. You must define at least one platform in\n" +
			"your code by using the #define command with one or\n" +
			"more of the following constants:\n\n" +

			"TI82, TI83, TI83_PLUS, TI83_PLUS_SE,\n" +
			"TI84_PLUS, TI84_PLUS_SE, TI85, TI86\n\n" +

			"(i.e. to compile for the TI-83+: #define TI83_PLUS)"
		); Message(
			"Compilation complete. Save any output code generated\n" +
			"as z80 assembly source files, and then use an assembly\n" +
			"compiler to compile them. DO NOT ATTEMPT TO COMPILE\n" +
			"AND RUN PROGRAMS THAT HAD ERRORS IN THEM, they\n" +
			"are NOT SAFE and will crash your calc (and might destroy\n" +
			"other data that you have on there) if you do it anyway."
		); in = null;
	}

	public static void main(String[] args) {
		javax.swing.JFrame win = new javax.swing.JFrame(
			"SquirrelBox v1.4"
		); win.setDefaultCloseOperation(
			javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE
		); final HLz80compiler compiler = new HLz80compiler();
		compiler.init(win.getRootPane(), true);
		win.setIconImage(compiler.icon.getImage());
		win.addWindowListener(new java.awt.event.WindowAdapter() {
			public void windowClosing(java.awt.event.WindowEvent w) {
				compiler.exit();
			}
			public void windowClosed(java.awt.event.WindowEvent w) {
				compiler.exit();
			}
		}); win.setBounds(75,50,650,550);
		win.setVisible(true);
		compiler.Message(
"(the) SquirrelBox compiler v1.5.\n" +
"Copyright (C) 2007 Dan Cook\n\n" +

"This program is free software; you can redistribute it and/or\n" +
"modify it under the terms of the GNU General Public License\n" +
"as published by the Free Software Foundation; either version 2\n" +
"of the License, or (at your option) any later version.\n\n" +

"This program is distributed in the hope that it will be useful,\n" +
"but WITHOUT ANY WARRANTY; without even the implied warranty of\n" +
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the\n" +
"GNU General Public License for more details.\n\n" +

"You should have received a copy of the GNU General Public License\n" +
"along with this program; if not, write to the Free Software\n" +
"Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA."
		);
	}
}